/* emHeader.c
**
** Data type for handling the header lines of an e-mail.
**
*/

#include "mrmacs.h"
#include "tools.h"

#include "emHeader.h"

static ehList_t emGenHeader (char *name, char *value)
{
    char *p;
    ehList_t res = (ehList_t) galloc (__FILE__, __LINE__,
				      sizeof(struct ehList_s) +
				      strlen (name) + strlen (value) + 2);
    res->next = res->inext = (ehList_t) NULL;
    p = (char *) res + sizeof(struct ehList_s);
    res->name = p; strcpy (p, name); p += strlen (p) + 1;
    res->value = p; strcpy (p, value);
    return res;
}

/* emHeader_ins (&headers, name, value)
**     fgt ein neues Header-Element (name, value) in die Header-Liste ein.
**     Um eine Suche zu ermglichen enthalten die Strukturen fr die einzelnen
**     Header-Elemente eine weitere Liste, in der die Elemente mit jeweils
**     (abgesehen von Gro- und Kleinschreibung) gleichem Namen stehen. Falls
**     bereits ein Header-Element mit dem gegebenen Namen existiert, dann wird
**     das neue Element auch an diese Liste angehngt.
*/
int emHeader_ins (emHeader_t *h, char *name, char *value)
{
    int res = 0;
    ehList_t eh, el;
    ifnull (h->l) {
	el = h->f = h->l = emGenHeader (name, value);
    } else {
	el = h->l->next = emGenHeader (name, value);
	h->l = h->l->next;
    }
    for (eh = h->f; noNULL(eh); eh = eh->next) {
	if (!lccmp (name, eh->name, -1) && (eh != el)) {
	    while (noNULL(eh->inext)) { eh = eh->inext; }
	    eh->inext = el; res = 1; break;
	}
    }
    return res;
}

/* l_unlink (&header_struct, element)
**     entfernt `element' aus der Header-Liste `header_struct'.
*/
static void l_unlink (emHeader_t *h, ehList_t el)
{
    ehList_t eh = h->f;
    if (eh == el) {
	/* Falls es sich gleich um das erste Element handelt, dann mu es
	** aus der Header-Struktur ausgehngt werden.
	*/
	h->f = eh->next;
	eh->next = (ehList_t) NULL;
	ifnull (h->f) { h->l = (ehList_t) NULL; }
    } else {
	/* Ansonsten wird es in der Restliste gesucht und ausgehngt. */
	while (noNULL(eh->next)) {
	    if (eh->next == el) {
		eh->next = el->next; el->next = (ehList_t) NULL; break;
	    }
	    eh = eh->next;
	}
    }
}

/* emHeader_del (&headers, header_name, header_index)
**     dient zum Entfernen von E-Mail-Headern aus der Header-Struktur.
**     Zur Identifikation eines Headers mu neben dem Namen auch eine Nummer
**     herhalten, da Header mit dem gleichen Namen mehrfach vorkommen drfen.
**     Diese Nummer ist `header_index':
**     - ist `header_index' >= 0, dann wird - bei gefundenem `header_name' in
**       der Header-Liste das Element gelscht, dessen Position `header_index'
**       entspricht;
**     - ist `header_index' < 0, dann werden smtliche Header zu dem Namen
**       `header_name' gelscht.
**     Das Funktionsergebnis von `emHeader_del' ist 0, falls ein (oder mehrere)
**     Element(e) entfernt wurde(n) und 1, falls kein Element zum entfernen
**     gefunden wurde.
*/
int emHeader_del (emHeader_t *h, char *name, int pos)
{
    int ix;
    ehList_t eh = h->f, ep;
    for (eh = h->f; noNULL(eh); eh = eh->next) {
	if (!lccmp (eh->name, name, -1)) { break; }
    }
    ifnull (eh) { return 1; }
    /* Wenn pos == 0 ist, dann mu das erste Element aus der EQ-Liste
    ** entfernt werden.
    */
    if (pos == 0) { l_unlink (h, eh); gfree (eh); return 0; }
    /* Wenn pos > 0 ist, dann mu das Element zunchst in der EQ-Liste
    ** gefunden und aus dieser entfernt werden.
    */
    if (pos > 0) {
	ep = eh->next; ix = 1;
	while (noNULL(ep) && ix < pos) { eh = ep; ep = eh->next; }
	/* Wenn `ep == NULL' ist, dann wurde das gesuchte Element nicht
	** gefunden.
	*/
	ifnull (ep) { return 1; }
	/* ep zeigt auf das zu lschende Element, eh auf das letzte Element
	** davor.
	*/
	eh->next = ep->next; l_unlink (h, ep); gfree (ep); return 0;
    }
    /* Wenn pos < 0 ist, dann mssen alle Elemente aus der Liste entfernt
    ** werden.
    */
    do {
	/* Nimm das nchste Element aus der EQ-Liste */
	ep = eh->inext;
	/* Entferne es aus der Suchliste */
	l_unlink (h, eh);
	/* und gib es frei */
	gfree (eh); eh = ep;
    } while (noNULL(eh));
    return 0;
}

/* emHeader_get (&headers, name, index)
**     sucht mittels `name' und `index' nach einem Header-Element in `headers'.
**     Wenn das Element gefunden wurde, dann wird es zurckgegeben; sonst wird
**     NULL zurckgegeben.
*/
ehList_t emHeader_get (emHeader_t *h, char *name, int pos)
{
    int ix;
    ehList_t eh;
    if (pos < 0) { return (ehList_t) NULL; }
    for (eh = h->f; noNULL(eh); eh = eh->next) {
	if (!lccmp (name, eh->name, -1)) { break; }
    }
    unlessnull (eh) {
	for (ix = 0; noNULL(eh) && ix < pos; eh = eh->inext, ++ix);
    }
    return (eh);
}

void emHeader_free (emHeader_t *h)
{
    ehList_t eh, er;
    for (eh = h->f; noNULL(eh); eh = er) {
	er = eh->next;
	gfree (eh);
    }
    h->f = h->l = (ehList_t) NULL;
}

int emHeader_chg (emHeader_t *h, char *name, int pos, char *value)
{
    ehList_t eh, el, new;
    ifnull (eh = emHeader_get (h, name, pos)) { return 1; }
    new = emGenHeader (name, value);
    /* Ersetze das Element in der Hauptliste */
    if (eh == h->f) {
	h->f = new;
	/* Hnge die Restlisten von eh in new ein. */
	new->next = eh->next; new->inext = eh->inext;
    } else {
	for (el = h->f; noNULL(el); el = el->next) {
	    if (el->next == eh) {
		new->next = eh->next; new->inext = eh->inext;
		el->next = new; break;
	    }
	}
    }
    if (eh == h->l) { h->l = eh; }
    /* Ersetze eh in der Liste der Elemente mit gleichem Namen durch new
    ** - falls es in so einer Liste steht. Da alle Elemente, die in einer
    ** Identittsliste vorkommen knnen, die sind, die auch in der Hauptliste
    ** vorkommen, reicht es aus, nachzuschauen, ob der `inext'-Zeiger eines
    ** der Elemente in der Hauptliste auf eh zeigt, und dieses ggf. durch
    ** einen Zeiger auf `new' zu ersetzen.
    */
    for (el = h->f; noNULL(el); el = el->next) {
	if (el->inext == eh) { el->inext = new; break; }
    }
    /* Zu guter letzt mu der Speicherplatz von eh freigegeben werden. */
    gfree (eh);
    return 0;
}
